@@ -114,6 +114,7 @@ struct btd_adapter {
 	guint scheduler_id;		/* Scheduler handle */
 
 	struct hci_dev dev;		/* hci info */
+	gboolean pairable;		/* pairable state */
 };
 
 static inline DBusMessage *invalid_args(DBusMessage *msg)
@@ -331,6 +332,12 @@ struct pending_auth_info *adapter_new_auth_request(struct btd_adapter *adapter,
 	return info;
 }
 
+static void dev_info_free(struct remote_dev_info *dev)
+{
+	g_free(dev->alias);
+	g_free(dev);
+}
+
 int pending_remote_name_cancel(struct btd_adapter *adapter)
 {
 	struct remote_dev_info *dev, match;
@@ -360,7 +367,7 @@ int pending_remote_name_cancel(struct btd_adapter *adapter)
 	}
 
 	/* free discovered devices list */
-	g_slist_foreach(adapter->found_devices, (GFunc) g_free, NULL);
+	g_slist_foreach(adapter->found_devices, (GFunc) dev_info_free, NULL);
 	g_slist_free(adapter->found_devices);
 	adapter->found_devices = NULL;
 
@@ -443,6 +450,44 @@ static struct bonding_request_info *bonding_request_new(DBusConnection *conn,
 	return bonding;
 }
 
+static int set_limited_discoverable(int dd, const uint8_t *cls, gboolean limited)
+{
+	uint32_t dev_class;
+	int num = (limited ? 2 : 1);
+	uint8_t lap[] = { 0x33, 0x8b, 0x9e, 0x00, 0x8b, 0x9e };
+	/*
+	 * 1: giac
+	 * 2: giac + liac
+	 */
+	if (hci_write_current_iac_lap(dd, num, lap, HCI_REQ_TIMEOUT) < 0) {
+		int err = errno;
+		error("Can't write current IAC LAP: %s(%d)",
+				strerror(err), err);
+		return -err;
+	}
+
+	if (limited) {
+		if (cls[1] & 0x20)
+			return 0; /* Already limited */
+
+		dev_class = (cls[2] << 16) | ((cls[1] | 0x20) << 8) | cls[0];
+	} else {
+		if (!(cls[1] & 0x20))
+			return 0; /* Already clear */
+
+		dev_class = (cls[2] << 16) | ((cls[1] & 0xdf) << 8) | cls[0];
+	}
+
+	if (hci_write_class_of_dev(dd, dev_class, HCI_REQ_TIMEOUT) < 0) {
+		int err = errno;
+		error("Can't write class of device: %s (%d)",
+							strerror(err), err);
+		return -err;
+	}
+
+	return 0;
+}
+
 static const char *mode2str(uint8_t mode)
 {
 	switch(mode) {
@@ -583,7 +628,7 @@ static DBusMessage *set_mode(DBusConnection *conn, DBusMessage *msg,
 				adapter->discov_timeout_id = 0;
 			}
 
-			if (!adapter->mode_sessions && !adapter->discov_timeout)
+			if (!adapter->mode_sessions && adapter->discov_timeout)
 				adapter_set_discov_timeout(adapter,
 						adapter->discov_timeout * 1000);
 		}
@@ -622,12 +667,40 @@ static DBusMessage *set_discoverable(DBusConnection *conn, DBusMessage *msg,
 	strmode = discoverable ? "discoverable" : "connectable";
 	mode = get_mode(&adapter->bdaddr, strmode);
 
+	if (mode == MODE_DISCOVERABLE && adapter->pairable)
+		mode = MODE_LIMITED;
+
 	if (mode == adapter->mode)
 		return dbus_message_new_method_return(msg);
 
 	return set_mode(conn, msg, mode, data);
 }
 
+static DBusMessage *set_pairable(DBusConnection *conn, DBusMessage *msg,
+				gboolean pairable, void *data)
+{
+	struct btd_adapter *adapter = data;
+
+	if (pairable == adapter->pairable)
+		goto done;
+
+	adapter->pairable = pairable;
+
+	write_device_pairable(&adapter->bdaddr, pairable);
+
+	emit_property_changed(connection, adapter->path,
+				ADAPTER_INTERFACE, "Pairable",
+				DBUS_TYPE_BOOLEAN, &pairable);
+
+	if (adapter->scan_mode & SCAN_INQUIRY) {
+		uint8_t mode = pairable ? MODE_LIMITED : MODE_DISCOVERABLE;
+		return set_mode(conn, msg, mode, data);
+	}
+
+done:
+	return dbus_message_new_method_return(msg);
+}
+
 static struct session_req *find_session(GSList *list, DBusMessage *msg)
 {
 	GSList *l;
@@ -700,6 +773,13 @@ static void session_free(struct session_req *req)
 	g_free(req);
 }
 
+static void session_owner_exit(DBusConnection *conn, void *user_data)
+{
+	struct session_req *req = user_data;
+
+	session_free(req);
+}
+
 static struct session_req *session_ref(struct session_req *req)
 {
 	req->refcount++;
@@ -755,14 +835,12 @@ static void confirm_mode_cb(struct agent *agent, DBusError *err, void *data)
 
 	if (err && dbus_error_is_set(err)) {
 		reply = dbus_message_new_error(req->msg, err->name, err->message);
-		dbus_connection_send(req->conn, reply, NULL);
-		dbus_message_unref(reply);
+		g_dbus_send_message(req->conn, reply);
 		goto cleanup;
 	}
 
 	reply = set_mode(req->conn, req->msg, req->mode, req->adapter);
-	dbus_connection_send(req->conn, reply, NULL);
-	dbus_message_unref(reply);
+	g_dbus_send_message(req->conn, reply);
 
 	if (!find_session(req->adapter->mode_sessions, req->msg))
 		goto cleanup;
@@ -823,10 +901,9 @@ static DBusMessage *set_discoverable_timeout(DBusConnection *conn,
 
 	path = dbus_message_get_path(msg);
 
-	dbus_connection_emit_property_changed(conn, path,
-					ADAPTER_INTERFACE,
-					"DiscoverableTimeout",
-					DBUS_TYPE_UINT32, &timeout);
+	emit_property_changed(conn, path,
+				ADAPTER_INTERFACE, "DiscoverableTimeout",
+				DBUS_TYPE_UINT32, &timeout);
 
 	return dbus_message_new_method_return(msg);
 }
@@ -902,10 +979,8 @@ static DBusMessage *set_name(DBusConnection *conn, DBusMessage *msg,
 
 	path = dbus_message_get_path(msg);
 
-	dbus_connection_emit_property_changed(conn, path,
-					ADAPTER_INTERFACE,
-					"Name", DBUS_TYPE_STRING,
-					&name);
+	emit_property_changed(conn, path, ADAPTER_INTERFACE, "Name",
+				DBUS_TYPE_STRING, &name);
 
 	return dbus_message_new_method_return(msg);
 }
@@ -919,10 +994,8 @@ static void reply_authentication_failure(struct bonding_request_info *bonding)
 			bonding->hci_status : HCI_AUTHENTICATION_FAILURE;
 
 	reply = new_authentication_return(bonding->msg, status);
-	if (reply) {
-		dbus_connection_send(bonding->conn, reply, NULL);
-		dbus_message_unref(reply);
-	}
+	if (reply)
+		g_dbus_send_message(bonding->conn, reply);
 }
 
 struct btd_device *adapter_find_device(struct btd_adapter *adapter, const char *dest)
@@ -947,6 +1020,7 @@ struct btd_device *adapter_create_device(DBusConnection *conn,
 				struct btd_adapter *adapter, const char *address)
 {
 	struct btd_device *device;
+	const char *path;
 
 	debug("adapter_create_device(%s)", address);
 
@@ -958,6 +1032,14 @@ struct btd_device *adapter_create_device(DBusConnection *conn,
 
 	adapter->devices = g_slist_append(adapter->devices, device);
 
+	path = device_get_path(device);
+	g_dbus_emit_signal(conn, adapter->path,
+			ADAPTER_INTERFACE, "DeviceCreated",
+			DBUS_TYPE_OBJECT_PATH, &path,
+			DBUS_TYPE_INVALID);
+
+	adapter_update_devices(adapter);
+
 	return device;
 }
 
@@ -1033,9 +1115,8 @@ static DBusMessage *remove_bonding(DBusConnection *conn, DBusMessage *msg,
 
 		const gchar *dev_path = device_get_path(device);
 
-		dbus_connection_emit_property_changed(conn, dev_path,
-					DEVICE_INTERFACE, "Paired",
-					DBUS_TYPE_BOOLEAN, &paired);
+		emit_property_changed(conn, dev_path, DEVICE_INTERFACE,
+					"Paired", DBUS_TYPE_BOOLEAN, &paired);
 	}
 
 proceed:
@@ -1063,17 +1144,15 @@ void adapter_remove_device(DBusConnection *conn, struct btd_adapter *adapter,
 	delete_entry(&adapter->bdaddr, "profiles", dstaddr);
 	adapter->devices = g_slist_remove(adapter->devices, device);
 
-	if (!device_is_temporary(device)) {
+	if (!device_is_temporary(device))
 		remove_bonding(conn, NULL, dstaddr, adapter);
 
-		g_dbus_emit_signal(conn, adapter->path,
-				ADAPTER_INTERFACE,
-				"DeviceRemoved",
-				DBUS_TYPE_OBJECT_PATH, &dev_path,
-				DBUS_TYPE_INVALID);
+	g_dbus_emit_signal(conn, adapter->path,
+			ADAPTER_INTERFACE, "DeviceRemoved",
+			DBUS_TYPE_OBJECT_PATH, &dev_path,
+			DBUS_TYPE_INVALID);
 
-		adapter_update_devices(adapter);
-	}
+	adapter_update_devices(adapter);
 
 	agent = device_get_agent(device);
 
@@ -1312,7 +1391,7 @@ static void cancel_bonding(struct btd_adapter *adapter, gboolean exited)
 		bonding->cancel = TRUE;
 }
 
-static void create_bond_req_exit(void *user_data)
+static void create_bond_req_exit(DBusConnection *conn, void *user_data)
 {
 	struct btd_adapter *adapter = user_data;
 
@@ -1519,7 +1598,7 @@ static DBusMessage *adapter_start_discovery(DBusConnection *conn,
 
 done:
 	req = create_session(adapter, conn, msg, 0,
-			(GDBusWatchFunction) session_free);
+				session_owner_exit);
 
 	adapter->disc_sessions = g_slist_append(adapter->disc_sessions, req);
 
@@ -1582,38 +1661,37 @@ static DBusMessage *get_properties(DBusConnection *conn,
 
 	/* Address */
 	property = srcaddr;
-	dbus_message_iter_append_dict_entry(&dict, "Address",
-						DBUS_TYPE_STRING, &property);
+	dict_append_entry(&dict, "Address", DBUS_TYPE_STRING, &property);
 
 	/* Name */
 	memset(str, 0, sizeof(str));
 	strncpy(str, (char *) adapter->dev.name, 248);
 	property = str;
 
-	dbus_message_iter_append_dict_entry(&dict, "Name",
-						DBUS_TYPE_STRING, &property);
+	dict_append_entry(&dict, "Name", DBUS_TYPE_STRING, &property);
 
 	/* Mode */
 	property = mode2str(adapter->mode);
 
-	dbus_message_iter_append_dict_entry(&dict, "Mode",
-						DBUS_TYPE_STRING, &property);
+	dict_append_entry(&dict, "Mode", DBUS_TYPE_STRING, &property);
 
 	/* Powered */
 	if (main_opts.offmode == HCID_OFFMODE_DEVDOWN)
 		value = adapter->up ? TRUE : FALSE;
 	else
 		value = adapter->scan_mode == SCAN_DISABLED ? FALSE : TRUE;
-	dbus_message_iter_append_dict_entry(&dict, "Powered",
-					DBUS_TYPE_BOOLEAN, &value);
+	dict_append_entry(&dict, "Powered", DBUS_TYPE_BOOLEAN, &value);
 
 	/* Discoverable */
 	value = adapter->scan_mode & SCAN_INQUIRY ? TRUE : FALSE;
-	dbus_message_iter_append_dict_entry(&dict, "Discoverable",
-					DBUS_TYPE_BOOLEAN, &value);
+	dict_append_entry(&dict, "Discoverable", DBUS_TYPE_BOOLEAN, &value);
+
+	/* Pairable */
+	dict_append_entry(&dict, "Pairable", DBUS_TYPE_BOOLEAN,
+				&adapter->pairable);
 
 	/* DiscoverableTimeout */
-	dbus_message_iter_append_dict_entry(&dict, "DiscoverableTimeout",
+	dict_append_entry(&dict, "DiscoverableTimeout",
 				DBUS_TYPE_UINT32, &adapter->discov_timeout);
 
 	if (adapter->state & PERIODIC_INQUIRY || adapter->state & STD_INQUIRY)
@@ -1622,8 +1700,7 @@ static DBusMessage *get_properties(DBusConnection *conn,
 		value = FALSE;
 
 	/* Discovering */
-	dbus_message_iter_append_dict_entry(&dict, "Discovering",
-					DBUS_TYPE_BOOLEAN, &value);
+	dict_append_entry(&dict, "Discovering", DBUS_TYPE_BOOLEAN, &value);
 
 	/* Devices */
 	devices = g_new0(char *, g_slist_length(adapter->devices) + 1);
@@ -1631,8 +1708,7 @@ static DBusMessage *get_properties(DBusConnection *conn,
 		struct btd_device *dev = l->data;
 		devices[i] = (char *) device_get_path(dev);
 	}
-	dbus_message_iter_append_dict_entry(&dict, "Devices",
-			DBUS_TYPE_ARRAY, &devices);
+	dict_append_array(&dict, "Devices", DBUS_TYPE_OBJECT_PATH, &devices, i);
 	g_free(devices);
 
 	dbus_message_iter_close_container(&iter, &dict);
@@ -1717,6 +1793,15 @@ static DBusMessage *set_property(DBusConnection *conn,
 		dbus_message_iter_get_basic(&sub, &timeout);
 
 		return set_discoverable_timeout(conn, msg, timeout, data);
+	} else if (g_str_equal("Pairable", property)) {
+		gboolean pairable;
+
+		if (dbus_message_iter_get_arg_type(&sub) != DBUS_TYPE_BOOLEAN)
+			return invalid_args(msg);
+
+		dbus_message_iter_get_basic(&sub, &pairable);
+
+		return set_pairable(conn, msg, pairable, data);
 	}
 
 	return invalid_args(msg);
@@ -1748,7 +1833,7 @@ static DBusMessage *mode_request(DBusConnection *conn,
 	req = find_session(adapter->mode_sessions, msg);
 	if (!req) {
 		req = create_session(adapter, conn, msg, new_mode,
-					(GDBusWatchFunction) session_free);
+					session_owner_exit);
 		adapter->mode_sessions = g_slist_append(adapter->mode_sessions,
 					req);
 	} else {
@@ -1895,7 +1980,7 @@ static DBusMessage *create_device(DBusConnection *conn,
 
 	debug("create_device(%s)", address);
 
-	device = device_create(conn, adapter, address);
+	device = adapter_create_device(conn, adapter, address);
 	if (!device)
 		return NULL;
 
@@ -1903,8 +1988,6 @@ static DBusMessage *create_device(DBusConnection *conn,
 
 	device_browse(device, conn, msg, NULL, FALSE);
 
-	adapter->devices = g_slist_append(adapter->devices, device);
-
 	return NULL;
 }
 
@@ -2368,6 +2451,10 @@ static void adapter_up(struct btd_adapter *adapter, int dd)
 	adapter->discov_timeout = get_discoverable_timeout(srcaddr);
 	adapter->state = DISCOVER_TYPE_NONE;
 
+	/* Set pairable mode */
+	if (read_device_pairable(&adapter->bdaddr, &adapter->pairable) < 0)
+		adapter->pairable = TRUE;
+
 	/* Set scan mode */
 	if (read_device_mode(srcaddr, mode, sizeof(mode)) == 0) {
 		if (!strcmp(mode, "off")) {
@@ -2396,7 +2483,8 @@ static void adapter_up(struct btd_adapter *adapter, int dd)
 		} else if (!strcmp(mode, "discoverable")) {
 			/* Set discoverable only if timeout is 0 */
 			if (adapter->discov_timeout == 0) {
-				adapter->mode = MODE_DISCOVERABLE;
+				adapter->mode = adapter->pairable ?
+					MODE_LIMITED : MODE_DISCOVERABLE;
 				adapter->scan_mode = SCAN_PAGE | SCAN_INQUIRY;
 			} else {
 				adapter->mode = MODE_CONNECTABLE;
@@ -2441,22 +2529,24 @@ static void adapter_up(struct btd_adapter *adapter, int dd)
 
 	pmode = mode2str(adapter->mode);
 
-	dbus_connection_emit_property_changed(connection, adapter->path,
-					ADAPTER_INTERFACE, "Mode",
-					DBUS_TYPE_STRING, &pmode);
+	emit_property_changed(connection, adapter->path, ADAPTER_INTERFACE,
+				"Mode", DBUS_TYPE_STRING, &pmode);
 
 	powered = adapter->scan_mode == SCAN_DISABLED ? FALSE : TRUE;
 
-	dbus_connection_emit_property_changed(connection, adapter->path,
-					ADAPTER_INTERFACE, "Powered",
-					DBUS_TYPE_BOOLEAN, &powered);
+	emit_property_changed(connection, adapter->path, ADAPTER_INTERFACE,
+				"Powered", DBUS_TYPE_BOOLEAN, &powered);
 
 	discoverable = adapter->scan_mode == (SCAN_PAGE | SCAN_INQUIRY) ? TRUE
 				: FALSE;
 
-	dbus_connection_emit_property_changed(connection, adapter->path,
-					ADAPTER_INTERFACE, "Discoverable",
-					DBUS_TYPE_BOOLEAN, &discoverable);
+	emit_property_changed(connection, adapter->path,
+				ADAPTER_INTERFACE, "Discoverable",
+				DBUS_TYPE_BOOLEAN, &discoverable);
+
+	emit_property_changed(connection, adapter->path,
+				ADAPTER_INTERFACE, "Pairable",
+				DBUS_TYPE_BOOLEAN, &adapter->pairable);
 
 	load_drivers(adapter);
 	load_devices(adapter);
@@ -2647,7 +2737,8 @@ int adapter_stop(struct btd_adapter *adapter)
 	}
 
 	if (adapter->found_devices) {
-		g_slist_foreach(adapter->found_devices, (GFunc) g_free, NULL);
+		g_slist_foreach(adapter->found_devices,
+				(GFunc) dev_info_free, NULL);
 		g_slist_free(adapter->found_devices);
 		adapter->found_devices = NULL;
 	}
@@ -2670,20 +2761,18 @@ int adapter_stop(struct btd_adapter *adapter)
 		adapter->active_conn = NULL;
 	}
 
-	dbus_connection_emit_property_changed(connection, adapter->path,
-					ADAPTER_INTERFACE, "Mode",
-					DBUS_TYPE_STRING, &mode);
+	emit_property_changed(connection, adapter->path, ADAPTER_INTERFACE,
+				"Mode", DBUS_TYPE_STRING, &mode);
 
 	powered = FALSE;
 
-	dbus_connection_emit_property_changed(connection, adapter->path,
-					ADAPTER_INTERFACE, "Powered",
-					DBUS_TYPE_BOOLEAN, &powered);
+	emit_property_changed(connection, adapter->path, ADAPTER_INTERFACE,
+				"Powered", DBUS_TYPE_BOOLEAN, &powered);
 
 	if (adapter->scan_mode == (SCAN_PAGE | SCAN_INQUIRY)) {
 		discoverable = FALSE;
 
-		dbus_connection_emit_property_changed(connection, adapter->path,
+		emit_property_changed(connection, adapter->path,
 					ADAPTER_INTERFACE, "Discoverable",
 					DBUS_TYPE_BOOLEAN, &discoverable);
 	}
@@ -2693,8 +2782,6 @@ int adapter_stop(struct btd_adapter *adapter)
 	adapter->mode = MODE_OFF;
 	adapter->state = DISCOVER_TYPE_NONE;
 
-	unload_drivers(adapter);
-
 	info("Adapter %s has been disabled", adapter->path);
 
 	return 0;
@@ -2771,11 +2858,12 @@ struct btd_adapter *adapter_create(DBusConnection *conn, int id)
 {
 	char path[MAX_PATH_LENGTH];
 	struct btd_adapter *adapter;
+	const char *base_path = manager_get_base_path();
 
 	if (!connection)
 		connection = conn;
 
-	snprintf(path, sizeof(path), "%s/hci%d", "/org/bluez", id);
+	snprintf(path, sizeof(path), "%s/hci%d", base_path, id);
 
 	adapter = g_try_new0(struct btd_adapter, 1);
 	if (!adapter) {
@@ -2806,6 +2894,8 @@ void adapter_remove(struct btd_adapter *adapter)
 
 	debug("Removing adapter %s", path);
 
+	unload_drivers(adapter);
+
 	for (l = adapter->devices; l; l = l->next)
 		device_remove(connection, l->data);
 	g_slist_free(adapter->devices);
@@ -2944,11 +3034,12 @@ void adapter_set_state(struct btd_adapter *adapter, int state)
 	if (state & PERIODIC_INQUIRY || state & STD_INQUIRY)
 		discov_active = TRUE;
 	else if (adapter->disc_sessions && main_opts.inqmode)
-		adapter->scheduler_id = g_timeout_add(main_opts.inqmode * 1000,
+		adapter->scheduler_id = g_timeout_add_seconds(main_opts.inqmode,
 				(GSourceFunc) start_inquiry, adapter);
 
 	if (!discov_active && adapter->found_devices) {
-		g_slist_foreach(adapter->found_devices, (GFunc) g_free, NULL);
+		g_slist_foreach(adapter->found_devices,
+					(GFunc) dev_info_free, NULL);
 		g_slist_free(adapter->found_devices);
 		adapter->found_devices = NULL;
 	}
@@ -2959,7 +3050,7 @@ void adapter_set_state(struct btd_adapter *adapter, int state)
 		adapter->oor_devices = NULL;
 	}
 
-	dbus_connection_emit_property_changed(connection, path,
+	emit_property_changed(connection, path,
 				ADAPTER_INTERFACE, "Discovering",
 				DBUS_TYPE_BOOLEAN, &discov_active);
 
@@ -2995,7 +3086,8 @@ int dev_rssi_cmp(struct remote_dev_info *d1, struct remote_dev_info *d2)
 }
 
 int adapter_add_found_device(struct btd_adapter *adapter, bdaddr_t *bdaddr,
-				int8_t rssi, name_status_t name_status)
+				int8_t rssi, uint32_t class, const char *alias,
+				name_status_t name_status)
 {
 	struct remote_dev_info *dev, match;
 
@@ -3010,6 +3102,13 @@ int adapter_add_found_device(struct btd_adapter *adapter, bdaddr_t *bdaddr,
 		if (rssi != 0)
 			dev->rssi = rssi;
 
+		dev->class = class;
+
+		if (alias) {
+			g_free(dev->alias);
+			dev->alias = g_strdup(alias);
+		}
+
 		 /* Get remote name can be received while inquiring.
 		  * Keep in mind that multiple inquiry result events can
 		  * be received from the same remote device.
@@ -3027,6 +3126,9 @@ int adapter_add_found_device(struct btd_adapter *adapter, bdaddr_t *bdaddr,
 
 	bacpy(&dev->bdaddr, bdaddr);
 	dev->rssi = rssi;
+	dev->class = class;
+	if (alias)
+		dev->alias = g_strdup(alias);
 	dev->name_status = name_status;
 
 	adapter->found_devices = g_slist_insert_sorted(adapter->found_devices,
@@ -3047,7 +3149,7 @@ int adapter_remove_found_device(struct btd_adapter *adapter, bdaddr_t *bdaddr)
 		return -1;
 
 	adapter->found_devices = g_slist_remove(adapter->found_devices, dev);
-	g_free(dev);
+	dev_info_free(dev);
 
 	return 0;
 }
@@ -3132,19 +3234,18 @@ void adapter_mode_changed(struct btd_adapter *adapter, uint8_t scan_mode)
 		return;
 	}
 
-	dbus_connection_emit_property_changed(connection, path,
-						ADAPTER_INTERFACE, "Mode",
-						DBUS_TYPE_STRING, &mode);
+	emit_property_changed(connection, path, ADAPTER_INTERFACE, "Mode",
+				DBUS_TYPE_STRING, &mode);
 
 	if (powered == FALSE || adapter->scan_mode == SCAN_DISABLED) {
-		dbus_connection_emit_property_changed(connection, path,
-						ADAPTER_INTERFACE, "Powered",
-						DBUS_TYPE_BOOLEAN, &powered);
+		emit_property_changed(connection, path,
+					ADAPTER_INTERFACE, "Powered",
+					DBUS_TYPE_BOOLEAN, &powered);
 	}
 
-	dbus_connection_emit_property_changed(connection, path,
-					ADAPTER_INTERFACE, "Discoverable",
-					DBUS_TYPE_BOOLEAN, &discoverable);
+	emit_property_changed(connection, path,
+				ADAPTER_INTERFACE, "Discoverable",
+				DBUS_TYPE_BOOLEAN, &discoverable);
 
 	adapter_set_scan_mode(adapter, scan_mode);
 }
@@ -3191,7 +3292,7 @@ struct active_conn_info *adapter_search_active_conn_by_bdaddr(struct btd_adapter
 	if (!adapter || !adapter->active_conn)
 		return NULL;
 
-	l = g_slist_find_custom(adapter->active_conn, &bda,
+	l = g_slist_find_custom(adapter->active_conn, bda,
 					active_conn_find_by_bdaddr);
 	if (l)
 		return l->data;
@@ -3246,10 +3347,29 @@ gboolean adapter_has_discov_sessions(struct btd_adapter *adapter)
 	return TRUE;
 }
 
+static void probe_driver(gpointer data, gpointer user_data)
+{
+	struct btd_adapter *adapter = data;
+	struct btd_adapter_driver *driver = user_data;
+
+	if (!adapter->up)
+		return;
+
+	driver->probe(adapter);
+}
+
 int btd_register_adapter_driver(struct btd_adapter_driver *driver)
 {
+	GSList *adapters;
+
 	adapter_drivers = g_slist_append(adapter_drivers, driver);
 
+	if (driver->probe == NULL)
+		return 0;
+
+	adapters = manager_get_adapters();
+	g_slist_foreach(adapters, probe_driver, driver);
+
 	return 0;
 }
 
@@ -3267,31 +3387,23 @@ static void agent_auth_cb(struct agent *agent, DBusError *derr, void *user_data)
 	g_free(auth);
 }
 
-int btd_request_authorization(const bdaddr_t *src, const bdaddr_t *dst,
+static int btd_adapter_authorize(struct btd_adapter *adapter, const bdaddr_t *dst,
 		const char *uuid, service_auth_cb cb, void *user_data)
 {
 	struct service_auth *auth;
-	struct btd_adapter *adapter;
 	struct btd_device *device;
 	struct agent *agent;
 	char address[18];
 	gboolean trusted;
 	const gchar *dev_path;
 
-	if (src == NULL || dst == NULL)
-		return -EINVAL;
-
-	adapter = manager_find_adapter(src);
-	if (!adapter)
-		return -EPERM;
-
 	/* Device connected? */
 	if (!g_slist_find_custom(adapter->active_conn,
 				dst, active_conn_find_by_bdaddr))
 		return -ENOTCONN;
 
 	ba2str(dst, address);
-	trusted = read_trust(src, address, GLOBAL_TRUST);
+	trusted = read_trust(&adapter->bdaddr, address, GLOBAL_TRUST);
 
 	if (trusted) {
 		cb(NULL, user_data);
@@ -3322,6 +3434,40 @@ int btd_request_authorization(const bdaddr_t *src, const bdaddr_t *dst,
 	return agent_authorize(agent, dev_path, uuid, agent_auth_cb, auth);
 }
 
+int btd_request_authorization(const bdaddr_t *src, const bdaddr_t *dst,
+		const char *uuid, service_auth_cb cb, void *user_data)
+{
+	struct btd_adapter *adapter;
+	GSList *adapters;
+
+	if (src == NULL || dst == NULL)
+		return -EINVAL;
+
+	if (bacmp(src, BDADDR_ANY) != 0)
+		goto proceed;
+
+	/* Handle request authorization for ANY adapter */
+	adapters = manager_get_adapters();
+
+	for (; adapters; adapters = adapters->next) {
+		int err;
+		adapter = adapters->data;
+
+		err = btd_adapter_authorize(adapter, dst, uuid, cb, user_data);
+		if (err == 0)
+			return 0;
+	}
+
+	return -EPERM;
+
+proceed:
+	adapter = manager_find_adapter(src);
+	if (!adapter)
+		return -EPERM;
+
+	return btd_adapter_authorize(adapter, dst, uuid, cb, user_data);
+}
+
 int btd_cancel_authorization(const bdaddr_t *src, const bdaddr_t *dst)
 {
 	struct btd_adapter *adapter = manager_find_adapter(src);
@@ -3366,8 +3512,51 @@ void adapter_update_devices(struct btd_adapter *adapter)
 		devices[i] = (char *) device_get_path(dev);
 	}
 
-	dbus_connection_emit_property_changed(connection, adapter->path,
+	emit_array_property_changed(connection, adapter->path,
 					ADAPTER_INTERFACE, "Devices",
-					DBUS_TYPE_ARRAY, &devices);
+					DBUS_TYPE_OBJECT_PATH, &devices);
 	g_free(devices);
 }
+
+static gchar *adapter_any_path = NULL;
+static int adapter_any_refcount = 0;
+
+const char *adapter_any_get_path(void)
+{
+	return adapter_any_path;
+}
+
+const char *btd_adapter_any_request_path(void)
+{
+	if (adapter_any_refcount > 0)
+		return adapter_any_path;
+
+	adapter_any_path = g_strdup_printf("%s/any", manager_get_base_path());
+	adapter_any_refcount++;
+
+	return adapter_any_path;
+}
+
+void btd_adapter_any_release_path(void)
+{
+	adapter_any_refcount--;
+
+	if (adapter_any_refcount > 0)
+		return;
+
+	g_free(adapter_any_path);
+	adapter_any_path = NULL;
+}
+
+gboolean adapter_is_pairable(struct btd_adapter *adapter)
+{
+	return adapter->pairable;
+}
+
+gboolean adapter_pairing_initiator(struct btd_adapter *adapter, bdaddr_t *bda)
+{
+	if (!adapter->bonding)
+		return FALSE;
+
+	return (bacmp(&adapter->bonding->bdaddr, bda) == 0);
+}
